# an order book manager for the world order book
# this will be improved by hands more skilled than mine, but i hope it will be useful until that day comes 

import logging as log

class Book: 
    """this class represents an order book for a single pair of tokens. it is initialized with the token pair and can be populated with orders. it can also be updated with new orders, order fills, and order removals. it is intended to be used as a data structure to facilitate the in memory representation of the world order book.

    :param token0: the address of the pay token
    :type token0: str
    :param token1: the address of the buy token
    :type token1: str
    """

    # the basic idea here is that we will keep a list of sorted order ids that map to a dictionary of order details

    # initialize the book
    def __init__(self, token0, token1):
        """constructor method"""

        # set the token pair
        self.token0 = token0
        self.token1 = token1

        # initialize the order list
        self.orders = []

        # initialize the order details dictionary
        self.details = {}

    # a function to populate the book with orders for the pair from the market
    def populate(self, book):
        """populates the book class with orders and order details 

        :param book: an array of order objects, in the format of [[token0_amount, token1_amount, order_id], ...]
        :type book: list
        """

        # iterate through the list of orders to populate the list and details dictionary
        for order in book[0]:

            # add the order id to the list 
            self.orders.append(order[2])

            # populate the details dictionary with the order details
            self.details[order[2]] = [order[0], order[1]]

    # a function to add a new order to the book 
    def add(self, order_id, token0_amt, token1_amt): 
        """adds a new order to the book class 

        :param order_id: the id of the order to add
        :type order_id: int
        :param token0_amt: the amount the order is willing to pay in token0
        :type token0_amt: int
        :param token1_amt: the amount the order is willing to pay in token1
        :type token1_amt: int
        :returns: True if the update was successful, False otherwise
        :rtype: bool
        """

        # add the order details to the dictionary 
        self.details[order_id] = [token0_amt, token1_amt]

        # now determine where to insert the order id in the sorted list 
        for i in range(len(self.orders)):

            # if the order is a better price insert it before the current order
            if ((token0_amt / token1_amt) * (self.details[self.orders[i]][1] / self.details[self.orders[i]][0])) >= 1:

                self.orders.insert(i, order_id)
                return True

        # if the order is the worst price insert it at the end of the list
        self.orders.append(order_id)

        log.info('adding order %s to book', order_id)

        return True

    # a function to update fill on an order
    def fill(self, order_id, token0_fill, token1_fill):
        """updates an order in the book class with a fill

        :param order_id: the id of the order to update
        :type order_id: int
        :param token0_fill: the amount of token0 that was payed 
        :type token0_fill: int
        :param token1_fill: the amount of token1 that was received 
        :type token1_fill: int
        :returns: True if the update was successful, False otherwise
        :rtype: bool
        """

        # update the order details
        try: 
            self.details[order_id][0] -= token0_fill
            self.details[order_id][1] -= token1_fill
        except Exception as e:
            log.error('unable to fill order %s', e)
            return False

        # if the order is filled remove it from the book
        if self.details[order_id][0] == 0 and self.details[order_id][1] == 0:

            try:
                self.remove(order_id, 0, 0)
            except Exception as e:
                log.error('unable to remove order %s after fill', e)
                return False

        log.info('filling order %s in book', order_id)

        return True

    # a function to remove an order from the book
    # TODO: the token amounts are not needed here, but they are currently here to ensure consistency with the other functions and data structures in the repo
    def remove(self, order_id, token0_amt, token1_amt):
        """removes an order from the book class

        :param order_id: the id of the order to remove
        :type order_id: int
        :param token0_amt: the pay amount remaining on the order
        :type token0_amt: int
        :param token1_amt: the buy amount remaining on the order
        :type token1_amt: int
        :returns: True if the update was successful, False otherwise
        :rtype: bool
        """

        # remove the order
        try:
            self.orders.remove(order_id)
            self.details.pop(order_id)
        except Exception as e:
            log.error('unable to remove order %s', e)
            return False

        log.info('removing order %s from book', order_id)

        return True

    # a function to take a parsed order and update the book accordingly
    # TODO: i think this could be better, specifically i think we should be able to pass the book object throughout the function calls to be updated
    # determining the structure of this book object in the larger context of the repo will be a problem to solve later
    # a current shortcoming of the market contract is offer deleted does not return any details on the pair, just the order id - because of this we are only checking to delted filled orders through checks on fill 
    def stream_update(self, parsed):
        """updates the book class with a parsed order object
        
        :param parsed: a parsed order object, each dictionary contains at the mininum the event type, order id, pay token amount, and buy token amount
        :type parsed: dict
        :returns: True if the update was successful, False otherwise
        :rtype: bool
        """

        # set the dictionary to determine the function to call based on the event type
        event_dict = {
            'LogMake' : self.add,
            'LogTake' : self.fill,
            'LogKill' : self.remove
        }
        
        try: 
            event_dict[parsed['event']](parsed['id'], parsed['pay_amt'], parsed['buy_amt'])
        except Exception as e:
            log.error('something has gone wrong in our stream update: %s', e)
            return False

        log.info('updating book with %s: %s', parsed['event'], parsed['id'])
        
        return True


from .router import RubiconRouter, RubiconRouterSigner
from .market import RubiconMarket, RubiconMarketSigner
from .house import BathHouse, BathHouseSigner
from .bath import BathToken, BathTokenSigner
from .aid import FactoryAid, FactoryAidSigner, MarketAid, MarketAidSigner





import os
import json

class OptimismMainnet:
    """this class represents a rolodex of relevant informatino regarding the Optimism Mainnet network, such as contract addresses, token addresses, and abis."""

    # TODO: set these to be dynamically pulled in a trustless fashion
    def __init__(self):
        self.network = 'Optimism'
        self.chain_id = 10
        self.currency = 'ETH'
        self.rpc_url = 'https://mainnet.optimism.io'
        self.explorer_url = 'https://optimistic.etherscan.io'
        
        # set the rubicon contract addresses
        self.market = '0x7a512d3609211e719737E82c7bb7271eC05Da70d'
        self.router = '0x7Af14ADc8Aea70f063c7eA3B2C1AD0D7A59C4bFf'
        self.house = '0x203328C161D23dCEee3E439DeEB25cA19e2c4984'
        self.pair = '0xF8780E00Ce8ed2e79aeC10908a169900eD1D4AFe'
        self.factory = '0x267D94C6e67e4436EFfE092b08d040cFF36B2DA7'

        # set the rubicon contract abis
        path = f"{os.path.dirname(os.path.realpath(__file__))}/abis/"
        self.market_abi = json.load(open(path + 'RubiconMarket.json'))
        self.router_abi = json.load(open(path + 'RubiconRouter.json'))
        self.house_abi = json.load(open(path + 'BathHouse.json'))
        self.pair_abi = json.load(open(path + 'BathPair.json'))
        self.bath_abi = json.load(open(path + 'BathToken.json'))
        self.aid_abi = json.load(open(path + 'MarketAid.json'))
        self.factory_abi = json.load(open(path + 'MarketAidFactory.json'))

        # TODO: for now we will hardcode the pool addresses, but in the future we will want to dynamically pull these in a trustless fashion 
            # this can be done through the house contract given the token addresses
        self.bathETH = '0xB0bE5d911E3BD4Ee2A8706cF1fAc8d767A550497'
        self.bathWBTC = '0x7571CC9895D8E997853B1e0A1521eBd8481aa186'
        self.bathUSDC = '0xe0e112e8f33d3f437D1F895cbb1A456836125952'
        self.bathDAI = '0x60daEC2Fc9d2e0de0577A5C708BcaDBA1458A833'
        self.bathUSDT = '0xfFBD695bf246c514110f5DAe3Fa88B8c2f42c411'
        self.bathSNX = '0xeb5F29AfaaA3f44eca8559c3e8173003060e919f'
        self.bathOP = '0x574a21fE5ea9666DbCA804C9d69d8Caf21d5322b'

        # set up some common asset addresses that may be used 
        # TODO: it would be nice to be able to dynamically pull in a token list in a trustless fashion and populate this list
        self.eth = '0xDeadDeAddeAddEAddeadDEaDDEAdDeaDDeAD0000'
        self.weth = '0x4200000000000000000000000000000000000006'
        self.wbtc = '0x68f180fcCe6836688e9084f035309E29Bf0A2095'
        self.usdc = '0x7F5c764cBc14f9669B88837ca1490cCa17c31607'
        self.dai = '0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1'
        self.usdt = '0x94b008aA00579c1307B0EF2c499aD98a8ce58e58'
        self.snx = '0x8700dAec35aF8Ff88c16BdF0418774CB3D7599B4'
        self.op = '0x4200000000000000000000000000000000000042'
        
    ######################################################################
    # sure there is plenty of helpful things we can do here in the future ;)
    ######################################################################

class OptimismGoerli:
    """this class represents a rolodex of relevant informatino regarding the Optimism Goerli network, such as contract addresses, token addresses, and abis."""

    # TODO: set these to be dynamically pulled in a trustless fashion
    def __init__(self):
        """Initialize the Optimism Mainnet class."""
        self.network = 'Optimism Goerli'
        self.chain_id = 420
        self.currency = 'ETH'
        self.rpc_url = 'https://goerli.optimism.io'
        self.explorer_url = 'https://goerli-explorer.optimism.io/'

        # set the rubicon contract addresses
        self.market = '0x6cD8666aBB003073e45D69E5b3aa0b0Fe9CDBF91'
        self.router = '0x6aaEd1985a0e011ca82BB5Df8ebd92063134fd7c'
        self.house = '0x1229036F63679B61910CB1463e5BB57f68D19bb2'
        self.pair = '0x9dBf17d518f722B5Aae5573D808B94024b635529'
        self.utility = '0xd282dB449cC64D136b9D9a4399E7e3F133472EaE'
        self.factory = '0x6838dd21aa01Bde8E600d499A95f9AE02f2bB376'

        # set the rubicon contract abis
        path = f"{os.path.dirname(os.path.realpath(__file__))}/abis/"
        self.market_abi = json.load(open(path + 'RubiconMarket.json'))
        self.router_abi = json.load(open(path + 'RubiconRouter.json'))
        self.house_abi = json.load(open(path + 'BathHouse.json'))
        self.pair_abi = json.load(open(path + 'BathPair.json'))
        self.bath_abi = json.load(open(path + 'BathToken.json'))
        self.aid_abi = json.load(open(path + 'MarketAid.json'))
        self.factory_abi = json.load(open(path + 'MarketAidFactory.json'))

        # set up some common asset addresses that may be used 
        # TODO: it would be nice to be able to dynamically pull in a token list in a trustless fashion and populate this list
        self.eth = '0xDeadDeAddeAddEAddeadDEaDDEAdDeaDDeAD0000'
        self.weth = '0x4200000000000000000000000000000000000006'
        self.op = '0xCeE7148028Ff1B08163343794E85883174a61393'
        self.usdc = '0x708394f89acd3E0644f774EA6c876BFACE70e600'

        ######################################################################
        # sure there is plenty of helpful things we can do here in the future ;)
        ######################################################################

# set a dictionary that maps the chain id to the network class
networks = {
    10: OptimismMainnet,
    420: OptimismGoerli
}


import hexbytes
import logging as log
from web3 import Web3 
from eth_abi.codec import ABICodec
from web3._utils.events import get_event_data

from rubi.contracts.helper import networks

class BathToken: 
    """this class represents the BathToken.sol contract and has read functionality for the contract
    
    :param w3: Web3 instance
    :type w3: Web3
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 object, optional
    """

    # init function
    def __init__(self, address, w3, contract=None):
        """constructor method"""

        chain = w3.eth.chain_id

        if contract:
            self.contract = contract
            self.address = self.contract.address
        else:  
            network = networks[chain]()
            self.contract = w3.eth.contract(address=address, abi=network.bath_abi)
            self.address = address

        # set the class variables
        self.chain = chain 
        self.w3 = w3
        self.log_deposit_abi = self.contract.events.LogDeposit()._get_event_abi()
        self.log_withdraw_abi = self.contract.events.LogWithdraw()._get_event_abi()
        self.transfer_abi = self.contract.events.Transfer()._get_event_abi()
        self.codec: ABICodec = w3.codec

    ######################################################################
    # read calls
    ######################################################################  

    # allowance(owner (address), spender (address))
    def allowance(self, owner, spender):
        """returns the allowance of the spender for the owner 

        :param owner: the owner of the tokens
        :type owner: str
        :param spender: the spender of the tokens
        :type spender: str
        :return: the allowance of the spender for the owner in the integer representation of the asset, None if there is an error
        :rtype: int, None
        """

        try: 
            allowance = self.contract.functions.allowance(owner, spender).call()
        except ValueError: 
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            allowance = self.contract.functions.allowance(Web3.to_checksum_address(owner), Web3.to_checksum_address(spender)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return allowance

    # asset()
    def asset(self):
        """returns the asset address of the contract

        :return: the asset address of the contract
        :rtype: str
        """

        try: 
            asset = self.contract.functions.asset().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return asset

    # balanceOf(account (address))
    def balanceOf(self, account):
        """returns the balance of the account in the integer representation of the asset

        :param account: the account to check the balance of
        :type account: str
        :return: the balance of the account in the integer representation of the asset, None if there is an error
        :rtype: int, None
        """

        try: 
            balance = self.contract.functions.balanceOf(account).call()
        except ValueError:  
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            balance = self.contract.functions.balanceOf(Web3.to_checksum_address(account)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return balance

    # convertToAssets(shares (uint256))
    def convertToAssets(self, shares):

        """converts shares to their claim on underlying assets amounts

        :param shares: the number of shares to convert to assets
        :type shares: int
        :return: the number of assets that the shares represent 
        :rtype: int
        """

        try: 
            assets = self.contract.functions.convertToAssets(shares).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return assets

    # convertToShares(assets (uint256))
    def convertToShares(self, assets):
        """convert assets to their potential share amounts

        :param assets: the number of assets to convert to shares
        :type assets: int
        :return: the number of shares that the assets represent
        :rtype: int
        """

        try: 
            shares = self.contract.functions.convertToShares(assets).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return shares

    # decimals()
    def decimals(self):
        """returns the number of decimals of the asset

        :return: the number of decimals of the asset
        :rtype: int
        """

        try: 
            decimals = self.contract.functions.decimals().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return decimals

    # feeBPS()
    def feeBPS(self):
        """gets the withdrawal fee in basis points

        :return: the withdrawal fee in basis points
        :rtype: int
        """

        try: 
            feeBPS = self.contract.functions.feeBPS().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return feeBPS

    # feeTo()
    def feeTo(self):
        """gets the address that receives the withdrawal fee

        :return: the address that receives the withdrawal fee
        :rtype: str
        """

        try: 
            feeTo = self.contract.functions.feeTo().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return feeTo

    # initialized()
    def initialized(self):
        """check if the contract has been initialized
        
        :return: True if the contract has been initialized, False otherwise
        :rtype: bool
        """

        try: 
            initialized = self.contract.functions.initialized().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return initialized

    # outstandingAmount()
    def outstandingAmount(self):
        """gets the amount of the underlying asset that is outstanding on the books
        
        :return: the amount of the underlying asset that is outstanding on the books, in the integer representation of the asset
        :rtype: int
        """

        try: 
            outstandingAmount = self.contract.functions.outstandingAmount().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return outstandingAmount

    # underlyingBalance()
    def underlyingBalance(self):
        """gets the underlying balance of assets on the contract
        
        :return: the underlying balance of assets on the contract, in the integer representation of the asset
        :rtype: int
        """

        try: 
            underlyingBalance = self.contract.functions.underlyingBalance().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return underlyingBalance

    # totalSupply()
    def totalSupply(self):
        """gets the total supply of shares
        
        :return: the total supply of shares
        :rtype: int
        """

        try: 
            totalSupply = self.contract.functions.totalSupply().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return totalSupply
    
    ######################################################################
    # events & helpers
    ######################################################################

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_transfer_hash(self):
        return self.w3.keccak(text="Transfer(address,address,uint256)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_transfer(self, data): 

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.transfer_abi, data['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            transfer = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'from': event['args']['from'],  
                    'to': event['args']['to'],
                    'value': event['args']['value']
            }
            return transfer

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_deposit(self, log): 

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.transfer_abi, log['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            transfer = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'from': event['args']['from'],  
                    'to': event['args']['to'],
                    'value': event['args']['value']
            }
            return transfer

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_deposit_hash(self):
        return self.w3.keccak(text="LogDeposit(uint256,IERC20,uint256,address,uint256,uint256,uint256)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_deposit(self, data): 

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_deposit_abi, data['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            deposit = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'deposit_amt': event['args']['depositedAmt'],  
                    'asset': event['args']['asset'],
                    'shares_received': event['args']['sharesReceived'],
                    'depositor': event['args']['depositor'],
                    'underlying_balance': event['args']['underlyingBalance'],
                    'outstanding_amount': event['args']['outstandingAmount'],
                    'total_supply': event['args']['totalSupply']
            }
            return deposit

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_deposit(self, log): 

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_deposit_abi, log['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            deposit = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'deposit_amt': event['args']['depositedAmt'],  
                    'asset': event['args']['asset'],
                    'shares_received': event['args']['sharesReceived'],
                    'depositor': event['args']['depositor'],
                    'underlying_balance': event['args']['underlyingBalance'],
                    'outstanding_amount': event['args']['outstandingAmount'],
                    'total_supply': event['args']['totalSupply']
            }
            return deposit

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_withdraw_hash(self):
        return self.w3.keccak(text="LogWithdraw(uint256,IERC20,uint256,address,uint256,address,uint256,uint256,uint256)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_withdraw(self, data): 

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_withdraw_abi, data['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            withdraw = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'amt_withdrawn': event['args']['amountWithdrawn'],  
                    'asset': event['args']['asset'],
                    'shares_withdrawn': event['args']['sharesWithdrawn'],
                    'withdrawer': event['args']['withdrawer'],
                    'fee': event['args']['fee'],
                    'fee_to': event['args']['feeTo'],
                    'underlying_balance': event['args']['underlyingBalance'],
                    'outstanding_amount': event['args']['outstandingAmount'],
                    'total_supply': event['args']['totalSupply']
            }
            return withdraw

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_withdraw(self, log): 

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_withdraw_abi, log['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            withdraw = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'amt_withdrawn': event['args']['amountWithdrawn'],  
                    'asset': event['args']['asset'],
                    'shares_withdrawn': event['args']['sharesWithdrawn'],
                    'withdrawer': event['args']['withdrawer'],
                    'fee': event['args']['fee'],
                    'fee_to': event['args']['feeTo'],
                    'underlying_balance': event['args']['underlyingBalance'],
                    'outstanding_amount': event['args']['outstandingAmount'],
                    'total_supply': event['args']['totalSupply']
            }
            return withdraw

        except Exception as e:
            log.error(e, exc_info=True)
            return None

class BathTokenSigner(BathToken):
    """this class represents the BathToken.sol contract and is the super class of the BathToken class. it has both read and write functionality, and inherits its read functionality from the BahtToken class

    :param address: the addres of the bathtoken contract instance to interact with
    :type address: str
    :param w3: the web3 instance to use for interacting with the contract
    :type w3: Web3 object
    :param wallet: the signers wallet address
    :type wallet: str
    :param key: the signers private key
    :type key: str
    :param contract: an optional parameter that allows you to pass in a contract instance, if none the contract will be instantiated from the rolodex.py file
    :type contract_address: str, optional
    """

    def __init__(self, address, w3, wallet, key, contract=None):
        super().__init__(address, w3, contract)
        self.wallet = wallet
        self.key = key

    ######################################################################
    # write calls
    ######################################################################

    # approve(spender (address), amount (uint256))
    def approve(self, spender, amount, nonce=None, gas=3000000, gas_price=None):
        
        """this function approves the spender to spend the amount of tokens on behalf of the signer

        :param spender: the address of the spender
        :type spender: str
        :param amount: the amount of tokens to approve the spender to spend
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the approve transaction, or None if the transaction failed
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try: 
            approve = self.contract.functions.approve(spender, amount).buildTransaction(txn)
            approve = self.w3.eth.account.sign_transaction(approve, private_key=self.key)
            self.w3.eth.send_raw_transaction(approve.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            approve = self.contract.functions.approve(self.w3.to_checksum_address(spender), amount).buildTransaction(txn)
            approve = self.w3.eth.account.sign_transaction(approve, private_key=self.key)
            self.w3.eth.send_raw_transaction(approve.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return approve

    # transfer(recipient (address), amount (uint256))
    def transfer(self, recipient, amount, nonce=None, gas=3000000, gas_price=None):
        """a function to transfer tokens from the signer to the recipient

        :param recipient: the address of the recipient
        :type recipient: str
        :param amount: the amount of tokens to transfer
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the transfer transaction, or None if the transaction failed
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            transfer = self.contract.functions.transfer(recipient, amount).buildTransaction(txn)
            transfer = self.w3.eth.account.sign_transaction(transfer, private_key=self.key)
            self.w3.eth.send_raw_transaction(transfer.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            transfer = self.contract.functions.transfer(self.w3.to_checksum_address(recipient), amount).buildTransaction(txn)
            transfer = self.w3.eth.account.sign_transaction(transfer, private_key=self.key)
            self.w3.eth.send_raw_transaction(transfer.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return transfer

    # transferFrom(sender (address), recipient (address), amount (uint256))
    def transfer_from(self, sender, recipient, amount, nonce=None, gas=3000000, gas_price=None):

        """a function to transfer tokens from the sender to the recipient on behalf of the signer

        :param sender: the address of the sender
        :type sender: str
        :param recipient: the address of the recipient
        :type recipient: str
        :param amount: the amount of tokens to transfer, in the integer representation of the token
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the transferFrom transaction, or None if the transaction failed
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)
    
        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            transfer_from = self.contract.functions.transferFrom(sender, recipient, amount).buildTransaction(txn)
            transfer_from = self.w3.eth.account.sign_transaction(transfer_from, private_key=self.key)
            self.w3.eth.send_raw_transaction(transfer_from.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            transfer_from = self.contract.functions.transferFrom(self.w3.to_checksum_address(sender), self.w3.to_checksum_address(recipient), amount).buildTransaction(txn)
            transfer_from = self.w3.eth.account.sign_transaction(transfer_from, private_key=self.key)
            self.w3.eth.send_raw_transaction(transfer_from.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return transfer_from

    # deposit(assets (uint256), receiver (address))
    def deposit(self, assets, receiver, nonce=None, gas=3000000, gas_price=None):
        """a function to deposit assets into the bathtoken contract and receive tokens in return
        
        :param assets: the amount of assets to deposit, in the integer representation of the token
        :type assets: int
        :param receiver: the address of the receiver of the tokens
        :type receiver: str
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            deposit = self.contract.functions.deposit(assets, receiver).buildTransaction(txn)
            deposit = self.w3.eth.account.sign_transaction(deposit, private_key=self.key)
            self.w3.eth.send_raw_transaction(deposit.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            deposit = self.contract.functions.deposit(assets, self.w3.to_checksum_address(receiver)).buildTransaction(txn)
            deposit = self.w3.eth.account.sign_transaction(deposit, private_key=self.key)
            self.w3.eth.send_raw_transaction(deposit.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

    # withdraw(shares (uint256))
    def withdraw(self, shares, nonce=None, gas=3000000, gas_price=None):
        """a function to withdraw the underlying asset from the bathtoken contract in exchange for tokens

        :param shares: the amount of tokens to burn, in the integer representation of the token
        :type shares: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the withdraw transaction, or None if the transaction failed
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)
        
        if gas_price is None:
            gas_price = self.w3.eth.gas_price
        
        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}
    
        try:
            withdraw = self.contract.functions.withdraw(shares).buildTransaction(txn)
            withdraw = self.w3.eth.account.sign_transaction(withdraw, private_key=self.key)
            self.w3.eth.send_raw_transaction(withdraw.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return withdraw

    # getAllBonusTokenReward()
    def get_all_bonus_token_reward(self, nonce=None, gas=3000000, gas_price=None):
        """a function to get all the bonus tokens that have been earned by the user

        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the getAllBonusTokenReward transaction, or None if the transaction failed
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)
        
        # TODO: figure out a way to dynamically set gas
        if gas is None:
            gas = 3000000
        
        if gas_price is None:
            gas_price = self.w3.eth.gas_price
        
        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            get_all_bonus_token_reward = self.contract.functions.getAllBonusTokenReward().buildTransaction(txn)
            get_all_bonus_token_reward = self.w3.eth.account.sign_transaction(get_all_bonus_token_reward, private_key=self.key)
            self.w3.eth.send_raw_transaction(get_all_bonus_token_reward.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return get_all_bonus_token_reward

    ######################################################################
    # 
    ######################################################################


import os
import json
import logging as log
from web3 import Web3 

class ERC20:
    """this class represents a contract that implements the ERC20 standard. it is used to read the contract instance. 

    :param w3: Web3 instance
    :type w3: Web3
    :param address: address of the contract
    :type address: str
    """

    # init function
    def __init__(self, w3, address, contract=None):
        """constructor method"""

        if contract:
            self.contract = contract
            self.address = contract.address
        else:
            path = f"{os.path.dirname(os.path.realpath(__file__))}/abis/"
            abi = json.load(open(path + 'ERC20.json'))

            try: 
                contract = w3.eth.contract(address=address, abi=abi)
            except ValueError:
                log.warning('most likely a checksum error... retrying with checksummed addresses')
                address = w3.to_checksum_address(address)
                contract = w3.eth.contract(address=address, abi=abi)
            except Exception as e:
                log.error(e, exc_info=True)
                return None

            # set the class variables
            self.contract = contract
            self.address = address

        self.w3 = w3

        # TODO: i believe decimals is an optional function that may not be implemented, so this may need an error handler
        self.decimal = self.contract.functions.decimals().call()

    ######################################################################
    # read calls
    ######################################################################  

    # allowance(owner (address), spender (address))
    def allowance(self, owner, spender):
        """reads the allowance of the spender from the owner for the erc20 contract

        :param owner: address that owns the erc20 tokens
        :type owner: str
        :param spender: address that is allowed to spend the erc20 tokens
        :type spender: str
        :return: the allowance of the spender from the owner for the contract, in the integer representation of the token
        :rtype: int
        """

        # TODO: see if there is a way to make the exception more specific to only the checksum error 
        try: 
            allowance = self.contract.functions.allowance(owner, spender).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            owner = self.w3.to_checksum_address(owner)
            spender = self.w3.to_checksum_address(spender)
            allowance = self.contract.functions.allowance(owner, spender).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return allowance

    # balanceOf(account (address))
    def balance_of(self, account):
        """reads the erc20 balance of the account 

        :param account: the address of the account to read the balance of
        :type account: str
        :return: the balance of the account, in the integer representation of the token
        :rtype: int
        """

        try: 
            balance = self.contract.functions.balanceOf(account).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            account = self.w3.to_checksum_address(account)
            balance = self.contract.functions.balanceOf(account).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return balance

    # totalSupply()
    def total_supply(self):
        """reads the total supply of the erc20 token
        
        :return: the total supply of the erc20 token, in the integer representation of the token
        :rtype: int
        """

        try: 
            total_supply = self.contract.functions.totalSupply().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return total_supply
    
    # decimals() 
    def decimals(self):
        """reads the number of decimals of the erc20 token - warning this is not a constant function, so it may result in an error in its current implementation

        :return: the number of decimals of the erc20 token
        :rtype: int
        """

        try: 
            decimals = self.contract.functions.decimals().call()
        except Exception as e:
            log.warning('error message: ', e)
            return None
        
        return decimals

    # name()
    def name(self):
        """reads the name of the erc20 token

        :return: the name of the erc20 token
        :rtype: str
        """
            
        try: 
            name = self.contract.functions.name().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return name

    # symbol()
    def symbol(self):
        """reads the symbol of the erc20 token

        :return: the symbol of the erc20 token
        :rtype: str
        """

        try: 
            symbol = self.contract.functions.symbol().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return symbol

class ERC20Signer(ERC20): 
    """this class represents a contract that implements the ERC20 standard. it is a super class of the ERC20 class as it has write functionality. it is used to read and write to the contract instance.
    
    :param w3: Web3 instance
    :type w3: Web3
    :param address: address of the erc20 contract
    :type address: str
    :param wallet: wallet address of the signer
    :type wallet: str
    :param key: private key of the signer
    :type key: str
    """

    def __init__(self, w3, address, wallet, key, contract=None):
        super().__init__(w3, address)
        self.chain = w3.eth.chain_id
        self.wallet = wallet
        self.key = key

    ######################################################################
    # write calls
    ######################################################################

    # approve(spender (address), amount (uint256))
    def approve(self, spender, amount, nonce=None, gas=3000000, gas_price=None):
        """approves the spender to spend the amount of the erc20 token from the signer's wallet

        :param spender: address of the spender
        :type spender: str
        :param amount: amount of the erc20 token to approve the spender to spend
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :return: the transaction object of the approve transaction
        :rtype: dict
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            # TODO: issue #7
            approve = self.contract.functions.approve(spender, amount).build_transaction(txn) 
            approve = self.w3.eth.account.sign_transaction(approve, self.key)
            self.w3.eth.send_raw_transaction(approve.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            approve = self.contract.functions.approve(self.w3.to_checksum_address(spender), amount).build_transaction(txn)
            approve = self.w3.eth.account.sign_transaction(approve, self.key)
            self.w3.eth.send_raw_transaction(approve.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return approve

    # transfer(recipient (address), amount (uint256))
    def transfer(self, recipient, amount, nonce=None, gas=3000000, gas_price=None):
        """transfers the amount of the erc20 token to the recipient

        :param recipient: address of the recipient
        :type recipient: str
        :param amount: amount of the erc20 token to transfer
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the transfer transaction
        :rtype: dict
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            transfer = self.contract.functions.transfer(recipient, amount).build_transaction(txn)
            transfer = self.w3.eth.account.sign_transaction(transfer, self.key)
            self.w3.eth.send_raw_transaction(transfer.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            transfer = self.contract.functions.transfer(self.w3.to_checksum_address(recipient), amount).build_transaction(txn)
            transfer = self.w3.eth.account.sign_transaction(transfer, self.key)
            self.w3.eth.send_raw_transaction(transfer.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return transfer

    # transferFrom(sender (address), recipient (address), amount (uint256))
    def transfer_from(self, sender, recipient, amount, nonce=None, gas=3000000, gas_price=None):
        """transfers the amount of the erc20 token from the sender to the recipient

        :param sender: address of the sender
        :type sender: str
        :param recipient: address of the recipient
        :type recipient: str
        :param amount: amount of the erc20 token to transfer
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the transferFrom transaction
        :rtype: dict
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price
        
        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            transfer_from = self.contract.functions.transferFrom(sender, recipient, amount).build_transaction(txn)
            transfer_from = self.w3.eth.account.sign_transaction(transfer_from, self.key)
            self.w3.eth.send_raw_transaction(transfer_from.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            transfer_from = self.contract.functions.transferFrom(self.w3.to_checksum_address(sender), self.w3.to_checksum_address(recipient), amount).build_transaction(txn)
            transfer_from = self.w3.eth.account.sign_transaction(transfer_from, self.key)
            self.w3.eth.send_raw_transaction(transfer_from.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return transfer_from
        

    # increaseAllowance(spender (address), addedValue (uint256))
    def increase_allowance(self, spender, added_value, nonce=None, gas=3000000, gas_price=None):
        """increases the allowance of the spender by the added_value

        :param spender: address of the spender
        :type spender: str
        :param added_value: amount to increase the allowance by, in the integer representation of the erc20 token
        :type added_value: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the increaseAllowance transaction
        :rtype: dict
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            increase_allowance = self.contract.functions.increaseAllowance(spender, added_value).build_transaction(txn)
            increase_allowance = self.w3.eth.account.sign_transaction(increase_allowance, self.key)
            self.w3.eth.send_raw_transaction(increase_allowance.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            increase_allowance = self.contract.functions.increaseAllowance(self.w3.to_checksum_address(spender), added_value).build_transaction(txn)
            increase_allowance = self.w3.eth.account.sign_transaction(increase_allowance, self.key)
            self.w3.eth.send_raw_transaction(increase_allowance.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return increase_allowance

    # decreaseAllowance(spender (address), subtractedValue (uint256))
    def decrease_allowance(self, spender, subtracted_value, nonce=None, gas=3000000, gas_price=None):
        """decreases the allowance of the spender by the subtracted_value

        :param spender: address of the spender
        :type spender: str
        :param subtracted_value: amount to decrease the allowance by, in the integer representation of the erc20 token
        :type subtracted_value: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the decreaseAllowance transaction
        :rtype: dict
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price
        
        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}
        
        try:
            decrease_allowance = self.contract.functions.decreaseAllowance(spender, subtracted_value).build_transaction(txn)
            decrease_allowance = self.w3.eth.account.sign_transaction(decrease_allowance, self.key)
            self.w3.eth.send_raw_transaction(decrease_allowance.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            decrease_allowance = self.contract.functions.decreaseAllowance(self.w3.to_checksum_address(spender), subtracted_value).build_transaction(txn)
            decrease_allowance = self.w3.eth.account.sign_transaction(decrease_allowance, self.key)
            self.w3.eth.send_raw_transaction(decrease_allowance.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return decrease_allowance

    ######################################################################
    # 
    ###################################################################### 


import hexbytes
import logging as log
from eth_abi import decode
from eth_abi.codec import ABICodec
from web3._utils.events import get_event_data

from rubi.contracts.helper import networks

class RubiconMarket: 
    """this class represents the RubiconMarket.sol contract and has read functionality.

    :param w3: Web3 instance
    :type w3: Web3
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 object, optional
    """

    def __init__(self, w3, contract=None):
        """constructor method"""

        chain = w3.eth.chain_id

        if contract:
            self.contract = contract
            self.address = self.contract.address
        else:
            network = networks[chain]()
            self.contract = w3.eth.contract(address=network.market, abi=network.market_abi)
            self.address = network.market

        self.chain = chain
        self.w3 = w3
        self.log_make_abi = self.contract.events.LogMake._get_event_abi()
        self.log_take_abi = self.contract.events.LogTake._get_event_abi()
        self.log_kill_abi = self.contract.events.LogKill._get_event_abi()
        self.offer_deleted_abi = self.contract.events.OfferDeleted._get_event_abi()
        self.codec: ABICodec = w3.codec

    ######################################################################
    # read calls
    ######################################################################

    # getBestOffer(sell_gem (address), buy_gem(address))
    def get_best_offer(self, sell_gem, buy_gem):
        """returns the best offer for the given pair of tokens

        :param sell_gem: the address of the token being sold by the maker
        :type sell_gem: str
        :param buy_gem: the address of the token being bought by the maker
        :type buy_gem: str
        :return: the id of the best offer on the book, None if there is no offer on the book
        :rtype: int, None
        """

        try: 
            best_offer = self.contract.functions.getBestOffer(sell_gem, buy_gem).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            sell_gem = self.w3.to_checksum_address(sell_gem)
            buy_gem = self.w3.to_checksum_address(buy_gem)
            best_offer = self.contract.functions.getBestOffer(sell_gem, buy_gem).call()
            # TODO: add error handling, local logging, and OT tracing
            # TODO: when you pass in two zero addresses, it returns zero, is this a bug or a feature?
                   # basically if you pass in any two addresses that don't have an offer, it returns zero
                   # the question now is, do we want to return zero or do we want to return an errror
                   # and handle this when the function then views the offer
                   # prolly the latter...
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return best_offer

    # getBetterOffer(id (uint256))
    def get_better_offer(self, id):
        """returns the id of the offer that is better than the given offer

        :param id: the id of the offer
        :type id: int
        :return: the id of the offer that is better than the given offer, none if there is no better offer
        :rtype: int, None
        """
            
        try: 
            better_offer = self.contract.functions.getBetterOffer(id).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return better_offer

    # getWorseOffer(id (uint256))
    def get_worse_offer(self, id):
        """returns the id of the offer that is worse than the given offer

        :param id: the id of the offer
        :type id: int
        :return: the id of the offer that is worse than the given offer, none if there is no worse offer
        :rtype: int, None
        """

        try: 
            worse_offer = self.contract.functions.getWorseOffer(id).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return worse_offer

    # getBuyAmount(buy_gem (address), pay_gem (address), pay_amt (uint256))
    def get_buy_amount(self, buy_gem, pay_gem, pay_amt):
        """returns the amount of buy_gem that can be bought with pay_amt of pay_gem

        :param buy_gem: the address of the token being bought
        :type buy_gem: str
        :param pay_gem: the address of the token being paid
        :type pay_gem: str
        :param pay_amt: the amount of pay_gem being paid, in the integer representation of the token amount
        :type pay_amt: int
        :return: the amount of buy_gem that can be bought with pay_amt of pay_gem, in the integer representation of the buy_gem amount
        :rtype: int
        """

        try: 
            buy_amount = self.contract.functions.getBuyAmount(buy_gem, pay_gem, pay_amt).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            pay_gem = self.w3.to_checksum_address(pay_gem)
            buy_gem = self.w3.to_checksum_address(buy_gem)
            buy_amount = self.contract.functions.getBuyAmount(buy_gem, pay_gem, pay_amt).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return buy_amount

    # getFeeBPS()
    def get_fee_bps(self):
        """returns the fee in basis points
        
        :return: the fee in basis points
        :rtype: int
        """

        try: 
            fee_bps = self.contract.functions.getFeeBPS().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return fee_bps

    # getOffer(id (uint256))
    # returns: [pay_amt (uint256), pay_gem (address), buy_amt (uint256), buy_gem (address)] - pay gem is what the offerer is selling, buy gem is what the offerer is buying
    def get_offer(self, id):
        """returns the offer with the given id in the form of a list - [pay_amt, pay_gem, buy_amt, buy_gem]

        :param id: the id of the offer
        :type id: int
        :return: the offer with the given id
        :rtype: array
        """

        try: 
            offer = self.contract.functions.getOffer(id).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return offer

    # getOfferCount(sell_gem (address), buy_gem (address))
    def get_offer_count(self, sell_gem, buy_gem):
        """returns the number of offers on the book for the given pair of tokens

        :param sell_gem: the address of the token being sold by the maker
        :type sell_gem: str
        :param buy_gem: the address of the token being bought by the maker
        :type buy_gem: str
        :return: the number of offers on the book for the given pair of tokens, None if there are no offers on the book for the given pair of tokens
        :rtype: int, None 
        """

        try: 
            offer_count = self.contract.functions.getOfferCount(sell_gem, buy_gem).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            sell_gem = self.w3.to_checksum_address(sell_gem)
            buy_gem = self.w3.to_checksum_address(buy_gem)
            offer_count = self.contract.functions.getOfferCount(sell_gem, buy_gem).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return offer_count

    # getOwner(id (uint256))
    def get_owner(self, id):
        """returns the address of the owner of the offer with the given id

        :param id: the id of the offer
        :type id: int
        :return: the address of the owner of the offer with the given id
        :rtype: str
        """

        try: 
            owner = self.contract.functions.getOwner(id).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return owner

    # getPayAmount(pay_gem (address), buy_gem (address), buy_amt (uint256))
    def get_pay_amount(self, pay_gem, buy_gem, buy_amt):
        """returns the amount of pay_gem that can be paid to buy buy_amt of buy_gem
        
        :param pay_gem: the address of the token being paid
        :type pay_gem: str
        :param buy_gem: the address of the token being bought
        :type buy_gem: str
        :param buy_amt: the amount of buy_gem being bought, in the integer representation of the token amount, returns None if the offer could not be filled
        :type buy_amt: int, None
        """

        try: 
            pay_amount = self.contract.functions.getPayAmount(pay_gem, buy_gem, buy_amt).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            pay_gem = self.w3.to_checksum_address(pay_gem)
            buy_gem = self.w3.to_checksum_address(buy_gem)
            pay_amount = self.contract.functions.getPayAmount(pay_gem, buy_gem, buy_amt).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return pay_amount

    # matchingEnabled()
    def matching_enabled(self):
        """returns whether or not matching is enabled, True if matching is enabled, False if matching is disabled

        :return: whether or not matching is enabled
        :rtype: bool
        """

        try: 
            matching_enabled = self.contract.functions.matchingEnabled().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return matching_enabled

    ######################################################################
    # events & helpers
    ######################################################################

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_make_hash(self):
        return self.w3.keccak(text="LogMake(bytes32,bytes32,address,address,address,uint128,uint128,uint64)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # i feel like this could be done much faster... 
    def stream_log_make(self, data): 

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_make_abi, data['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            offer = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'pay_gem': event['args']['pay_gem'],  
                    'buy_gem': event['args']['buy_gem'],
                    'pay_amt': event['args']['pay_amt'],
                    'buy_amt': event['args']['buy_amt'],
                    'timestamp': event['args']['timestamp'],
                    'owner': event['args']['maker']
            }
            return offer

        except Exception as e:
            log.error(e, exc_info=True)
            return None 

    def parse_log_make(self, log): 

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_make_abi, log)

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            offer = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'pay_gem': event['args']['pay_gem'],  
                    'buy_gem': event['args']['buy_gem'],
                    'pay_amt': event['args']['pay_amt'],
                    'buy_amt': event['args']['buy_amt'],
                    'timestamp': event['args']['timestamp'],
                    'owner': event['args']['maker']
            }
            return offer
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None 


    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_take_hash(self):
        return self.w3.keccak(text="LogTake(bytes32,bytes32,address,address,address,address,uint128,uint128,uint64)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_take(self, data):

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_take_abi, data['params']['result'])

            # decode the offer id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass the trade back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'pay_gem': event['args']['pay_gem'],
                    'buy_gem': event['args']['buy_gem'],
                    'pay_amt': event['args']['take_amt'],
                    'buy_amt': event['args']['give_amt'],
                    'timestamp': event['args']['timestamp'],
                    'maker': event['args']['maker'],
                    'taker': event['args']['taker']
            }
            return trade
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
    def parse_log_take(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_take_abi, log)

            # decode the offer id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass the trade back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'pay_gem': event['args']['pay_gem'],
                    'buy_gem': event['args']['buy_gem'],
                    'pay_amt': event['args']['take_amt'],
                    'buy_amt': event['args']['give_amt'],
                    'timestamp': event['args']['timestamp'],
                    'maker': event['args']['maker'],
                    'taker': event['args']['taker']
            }
            return trade
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_kill_hash(self): 
        return self.w3.keccak(text="LogKill(bytes32,bytes32,address,address,address,uint128,uint128,uint64)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_kill(self, data): 

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:  
            event = get_event_data(self.codec, self.log_kill_abi, data['params']['result'])

            # decode the offer id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass the killed trade back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            kill = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'pay_gem': event['args']['pay_gem'],
                    'buy_gem': event['args']['buy_gem'],
                    'pay_amt': event['args']['pay_amt'],
                    'buy_amt': event['args']['buy_amt'],
                    'timestamp': event['args']['timestamp'],
                    'maker': event['args']['maker']
            }
            return kill
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None
    
    def parse_log_kill(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_kill_abi, log)

            # decode the offer id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass the killed trade back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            kill = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'pay_gem': event['args']['pay_gem'],
                    'buy_gem': event['args']['buy_gem'],
                    'pay_amt': event['args']['pay_amt'],
                    'buy_amt': event['args']['buy_amt'],
                    'timestamp': event['args']['timestamp'],
                    'maker': event['args']['maker']
            }
            return kill
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_offer_deleted_hash(self): 
        return self.w3.keccak(text="OfferDeleted(bytes32)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_offer_deleted(self, data): 

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:  
            event = get_event_data(self.codec, self.offer_deleted_abi, data['params']['result'])

            # decode the offer id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass the deleted trade back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            deleted = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event']
            }
            return deleted
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None
    
    def parse_offer_deleted(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.offer_deleted_abi, log)

            # decode the offer id
            offer_id = decode(['uint256'], event['args']['id'])[0]

            # now pass the killed trade back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            deleted = {
                    'id': offer_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event']
            }
            return deleted
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None

class RubiconMarketSigner(RubiconMarket): 
    """this class represents the RubiconMarket.sol contract and is a super class of the RubiconMarket class. this class has read functionality and inherents the read functionality of the RubiconMarket class. 

    :param w3: a web3 instance
    :type w3: Web3
    :param wallet: the signers wallet address
    :type wallet: str
    :param key: the signers private key
    :type key: str
    :param contract: an optional parameter that allows you to pass in a contract instance, if none the contract will be instantiated from the rolodex.py file
    :type contract_address: str, optional
    """

    def __init__(self, w3, wallet, key, contract=None):
        super().__init__(w3, contract)
        self.wallet = wallet
        self.key = key

    ######################################################################
    # write calls
    ######################################################################

    # buy(id (uint256), amount (uint256))
    # the user must put in the amount of pay_gem that they want and will be charged the amount of buy_gem that the offer is asking for
    def buy(self, id, amount, nonce=None, gas=3000000, gas_price=None):
        """buy the amount of pay_gem from the offer with the id, in exchange for the amount of buy_gem at the price of the offer

        :param id: id of the offer
        :type id: int
        :param amount: amount of pay_gem to buy, in the integer representation of the token
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the buy transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}
        
        try:
            buy = self.contract.functions.buy(id, amount).build_transaction(txn)
            buy = self.w3.eth.account.sign_transaction(buy, self.key)
            self.w3.eth.send_raw_transaction(buy.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return buy  

    # buyAllAmount(buy_gem (address), buy_amt (uint256), pay_gem (address), max_fill_amount (uint256))
        # TODO: add clear explanation for function names and parameters
        # buy_gem is the token you want to buy
        # buy_amt is the amount of the token you want to buy
        # pay_gem is the token you want to pay with
        # max_fill is the maximum amount of the token you want to pay with
    def buy_all_amount(self, buy_gem, buy_amt, pay_gem, max_fill_amount, nonce=None, gas=3000000, gas_price=None):
        """buy the buy_amt of the buy_gem token in exchange for the pay_gem token only if it does not exceed the max_fill_amount of the pay_gem token
        
        :param buy_gem: address of the token you want to buy
        :type buy_gem: str
        :param buy_amt: amount of the token you want to buy, in the integer representation of the token
        :type buy_amt: int
        :param pay_gem: address of the token you want to pay with
        :type pay_gem: str
        :param max_fill_amount: maximum amount of the pay_gem token you want to pay with, in the integer representation of the token
        :type max_fill_amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the buyAllAmount transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try: 
            buy_all_amount = self.contract.functions.buyAllAmount(buy_gem, buy_amt, pay_gem, max_fill_amount).build_transaction(txn)
            buy_all_amount = self.w3.eth.account.sign_transaction(buy_all_amount, self.key)
            self.w3.eth.send_raw_transaction(buy_all_amount.rawTransaction)
        except ValueError: 
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            buy_all_amount = self.contract.functions.buyAllAmount(self.w3.to_checksum_address(buy_gem), buy_amt, self.w3.to_checksum_address(pay_gem), max_fill_amount).build_transaction(txn)
            buy_all_amount = self.w3.eth.account.sign_transaction(buy_all_amount, self.key)
            self.w3.eth.send_raw_transaction(buy_all_amount.rawTransaction)
        except Exception as e: 
            log.error(e, exc_info=True)
            return None
        
        return buy_all_amount

    # cancel(id (uint256))
    def cancel(self, id, nonce=None, gas=3000000, gas_price=None):
        """cancel the offer with the id, user can only cancel offers they have created

        :param id: id of the offer
        :type id: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the cancel transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            cancel = self.contract.functions.cancel(id).build_transaction(txn)
            cancel = self.w3.eth.account.sign_transaction(cancel, self.key)
            self.w3.eth.send_raw_transaction(cancel.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return cancel

    # offer(pay_amt (uint256), pay_gem (address), buy_amt (uint256), buy_gem (address))
    def offer(self, pay_amt, pay_gem, buy_amt, buy_gem, pos=0, nonce=None, gas=3000000, gas_price=None):
        """create an offer to buy the buy_amt of the buy_gem token in exchange for the pay_amt of the pay_gem token

        :param pay_amt: amount of the pay_gem token you want to pay with, in the integer representation of the token
        :type pay_amt: int
        :param pay_gem: address of the token you want to pay with
        :type pay_gem: str
        :param buy_amt: amount of the buy_gem token you want to buy, in the integer representation of the token
        :type buy_amt: int
        :param buy_gem: address of the token you want to buy
        :type buy_gem: str
        :param pos: position of the offer in the linked list, default to 0 unless the maker knows the position they want to insert the offer at
        :type pos: int, optional
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the offer transaction, returns None if the transaction fails
        :rtype: dict, None    
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            offer = self.contract.functions.offer(pay_amt, pay_gem, buy_amt, buy_gem, pos).build_transaction(txn)
            offer = self.w3.eth.account.sign_transaction(offer, self.key)
            self.w3.eth.send_raw_transaction(offer.rawTransaction)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            offer = self.contract.functions.offer(pay_amt, self.w3.to_checksum_address(pay_gem), buy_amt, self.w3.to_checksum_address(buy_gem), pos).build_transaction(txn)
            offer = self.w3.eth.account.sign_transaction(offer, self.key)
            self.w3.eth.send_raw_transaction(offer.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return offer

    # sellAllAmount(pay_gem (address), pay_amt (uint256), buy_gem (address), min_fill_amount (uint256))
    def sell_all_amount(self, pay_gem, pay_amt, buy_gem, min_fill_amount, nonce=None, gas=3000000, gas_price=None):
        """sell the pay_amt of the pay_gem token in exchange for buy_gem, on the condition that you receive at least the min_fill_amount of the buy_gem token

        :param pay_gem: address of the token you want to pay with
        :type pay_gem: str
        :param pay_amt: amount of the pay_gem token you want to pay with, in the integer representation of the token
        :type pay_amt: int
        :param buy_gem: address of the token you want to buy
        :type buy_gem: str
        :param min_fill_amount: minimum amount of the buy_gem token you want to receive, in the integer representation of the token
        :type min_fill_amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the sellAllAmount transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            sell_all_amount = self.contract.functions.sellAllAmount(pay_gem, pay_amt, buy_gem, min_fill_amount).build_transaction(txn)
            sell_all_amount = self.w3.eth.account.sign_transaction(sell_all_amount, self.key)
            self.w3.eth.send_raw_transaction(sell_all_amount.rawTransaction)
        except ValueError:
            print('most likely a checksum error... retrying with checksummed addresses')
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            sell_all_amount = self.contract.functions.sellAllAmount(self.w3.to_checksum_address(pay_gem), pay_amt, self.w3.to_checksum_address(buy_gem), min_fill_amount).build_transaction(txn)
            sell_all_amount = self.w3.eth.account.sign_transaction(sell_all_amount, self.key)
            self.w3.eth.send_raw_transaction(sell_all_amount.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return sell_all_amount


import hexbytes
import logging as log
from web3 import Web3 
from eth_abi import decode
from eth_abi.codec import ABICodec
from web3._utils.events import get_event_data

from rubi.contracts.helper import networks

class RubiconRouter: 
    """this class represents a RubiconRouter.sol contract with read functionality.
    
    :param w3: Web3 instance
    :type w3: Web3
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 Object, optional
    """

    # init function
    def __init__(self, w3, contract=None):

        chain = w3.eth.chain_id

        if contract:
            self.contract = contract
            self.address = self.contract.address
        else:
            network = networks[chain]()
            self.contract = w3.eth.contract(address=network.router, abi=network.router_abi)
            self.address = network.router

        # set the class variables
        self.w3 = w3
        self.log_swap_abi = self.contract.events.LogSwap._get_event_abi()
        self.codec: ABICodec = w3.codec

    ######################################################################
    # read calls
    ######################################################################  
    
    # checkClaimAllUserBonusTokens(user (address), targetBathTokens (address[]), token (address))
    def check_claim_all_user_bonus_tokens(self, user, targetBathTokens, token):
        """check the claim the user has to bonus tokens 

        :param user: the user address of interest
        :type user: str
        :param targetBathTokens: the bath tokens of interest, this is an array of addresses
        :type targetBathTokens: list
        :param token: the bonus token of interest
        :type token: str
        :return: the amount of bonus tokens the user has available to claim
        :rtype: int
        """

        try: 
            claimable = self.contract.functions.checkClaimAllUserBonusTokens(user, targetBathTokens, token).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return claimable

    # getBestOffer(asset (address), quote(address))
    def get_best_offer(self, asset, quote):
        """get the best offer for a given asset and quote pair
        
        :param asset: the asset address of interest
        :type asset: str
        :param quote: the quote address of interest
        :type quote: str
        :return: the best offer id for the given asset and quote pair
        :rtype: int
        """

        try: 
            best_offer = self.contract.functions.getBestOfferAndInfo(asset, quote).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            asset = self.w3.to_checksum_address(asset)
            quote = self.w3.to_checksum_address(quote)
            best_offer = self.contract.functions.getBestOfferAndInfo(asset, quote).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return best_offer

    # getBookFromPair(asset (address), quote(address), topNOrders (uint256))
    def get_book_from_pair(self, asset, quote, topNOrders):
        """get the book for a given asset and quote pair, to the depth of topNOrders
        
        :param asset: the asset address of interest
        :type asset: str
        :param quote: the quote address of interest
        :type quote: str
        :param topNOrders: the depth of the book to return
        :type topNOrders: int
        :return: the book for the given asset and quote pair, returned as a sorted list of [[asset_amount, quote_amount, offer_id], ...]]]
        :rtype: list
        """

        try: 
            book = self.contract.functions.getBookFromPair(asset, quote, topNOrders).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            asset = self.w3.to_checksum_address(asset)
            quote = self.w3.to_checksum_address(quote)
            book = self.contract.functions.getBookFromPair(asset, quote, topNOrders).call()
        except Exception as e:
            return None
        
        return book

    ######################################################################
    # events & helpers
    ######################################################################

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_swap_hash(self): 
        return self.w3.keccak(text="LogSwap(uint256,address,uint256,address,bytes32,uint256,address)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_swap(self, data): 

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:  
            event = get_event_data(self.codec, self.log_swap_abi, data['params']['result'])

            # TODO: parse out the pair from the bytes encoding that is passed in the event (bytes32)
            # now pass the swap back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            swap = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'inputAmount': event['args']['inputAmount'],
                    'inputERC20': event['args']['inputERC20'],
                    'hurdleBuyAmtMin': event['args']['hurdleBuyAmtMin'],
                    'targetERC20': event['args']['targetERC20'],
                    'pair': event['args']['pair'],
                    'realizedFill': event['args']['realizedFill'],
                    'recipient': event['args']['recipient'],
            }
            return swap
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None
    
    def parse_log_swap(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_swap_abi, log)

            # decode the pair 
            # TODO: this be broke 
            pair = decode(['bytes32'], event['args']['pair'])[0]

            # now pass the swap back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            swap = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'inputAmount': event['args']['inputAmount'],
                    'inputERC20': event['args']['inputERC20'],
                    'hurdleBuyAmtMin': event['args']['hurdleBuyAmtMin'],
                    'targetERC20': event['args']['targetERC20'],
                    'pair': pair,
                    'realizedFill': event['args']['realizedFill'],
                    'recipient': event['args']['recipient'],
            }
            return swap
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None

class RubiconRouterSigner(RubiconRouter):
    """this class represents a RubiconRouter.sol contract with read and write functionality. it is the super class of the RubiconRouter class, and inherints the read functionality from it.

    :param w3: Web3 instance
    :type w3: Web3
    :param wallet: the signers wallet address 
    :type wallet: str
    :param key: the signers private key
    :type key: str
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 Object, optional
    """

    def __init__(self, w3, wallet, key, contract=None):
        super().__init__(w3, contract)
        self.chain = w3.eth.chain_id
        self.wallet = wallet
        self.key = key

    ######################################################################
    # write calls
    ######################################################################

    # swap(pay_amt (uint256), buy_amt_min (uint256), route (address[]), expectedMarketFeeBPS (uint256))
    def swap(self, pay_amt, buy_amt_min, route, expected_market_fee_bps=1, nonce=None, gas=3000000, gas_price=None):
        """this function swaps the pay_amt of the first token in the route array for at least the buy_amt_min of the last token in the route array

        :param pay_amt: the amount of the first token in the route array to swap, as an integer amount of the token's smallest unit
        :type pay_amt: int
        :param buy_amt_min: the minimum amount of the last token in the route array to receive, as an integer amount of the token's smallest unit
        :type buy_amt_min: int
        :param route: array of addresses representing the path of tokens to swap through
        :type route: list
        :param expected_market_fee_bps: the market fee, defaults to 1
        :type expected_market_fee_bps: int, optional
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the swap transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try: 
            swap = self.contract.functions.swap(pay_amt, buy_amt_min, route, expected_market_fee_bps).build_transaction(txn)
            swap = self.w3.eth.account.sign_transaction(swap, self.key)
            self.w3.eth.send_raw_transaction(swap.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return swap


import json
import logging as log
from web3 import Web3
from attributedict.collections import AttributeDict

import rubi.contracts as contracts
from rubi.contracts.helper import networks

class Rubicon:
    """this class serves as a the main entry point to the repository. it acts as the initialization of multiple contract instances, and give access to the various functions of the contracts. it also provides a few helper functions to make interacting with the contracts easier. more to come soon!
    
    :param w3: a web3 instance
    :type w3: Web3
    :param wallet: the wallet address of the user, defaults to None
    :type wallet: str, optional
    :param key: the private key of the user, defaults to None
    :type key: str, optional
    :param market: a market contract object, defaults to None
    :type market: str, optional
    :param router: a router contract object, defaults to None
    :type router: str, optional
    :param factory: a factory contract object, defaults to None
    :type factory: str, optional
    """

    def __init__(self, w3, wallet=None, key=None, market=None, router=None, factory=None):

        # load required variables
        chain = w3.eth.chain_id

        try:
            network = networks[chain]()
        except:
            network = 'unknown'

        # set class variables 
        self.w3 = w3
        self.chain = chain
        self.network = network

        self.key = key
        self.wallet = wallet

        # if wallet and key are provided then create signer instances 
        if wallet and key:
            self.market = contracts.RubiconMarketSigner(w3, wallet, key, contract=market)
            self.router = contracts.RubiconRouterSigner(w3, wallet, key, contract=router)
            self.factory = contracts.FactoryAidSigner(w3, wallet, key, contract=factory)
        else:
            self.market = contracts.RubiconMarket(w3, contract=market)
            self.router = contracts.RubiconRouter(w3, contract=router)
            self.factory = contracts.FactoryAid(w3, contract=factory)

        # set the hashes of the event emits
        self.log_make_hash = self.market.get_log_make_hash()
        self.log_take_hash = self.market.get_log_take_hash()
        self.log_kill_hash = self.market.get_log_kill_hash()
        self.offer_deleted_hash = self.market.get_offer_deleted_hash()
        self.log_swap_hash = self.router.get_log_swap_hash()

        # create a dictionary to map market event emits to their respective functions
        self.market_events = {
            self.log_make_hash : self.market.stream_log_make,
            self.log_take_hash : self.market.stream_log_take,
            self.log_kill_hash : self.market.stream_log_kill
        }

    ######################################################################
    # read calls
    ######################################################################

    # a function that will identify the event topic and call the appropriate function
    def parse_market_events(self, event): 
        """a function to parse event emits from the market contract and call the appropriate function to handle the event

        :param event: the event emit from the market contract
        :type event: dict
        :return: the parsed event
        :rtype: dict
        """

        data = AttributeDict(json.loads(event))

        try: 
            parsed = self.market_events[data.params.result.topics[0]](data)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return parsed

    ######################################################################
    # write calls
    ######################################################################

    # a function to create and pass back a bath house instance
    def house(self, contract=None):
        """a function to create and pass back a bath house instance
        
        :param contract: a bath house contract instance, defaults to None
        :type contract: class, optional
        :return: a bath house instance
        :rtype: house.House
        """

        if self.wallet and self.key:
            return contracts.BathHouseSigner(self.w3, self.wallet, self.key, contract=contract)
        else:
            return contracts.BathHouse(self.w3, contract=contract)

    # a function to create and pass back a bath token instance
    def bath(self, token_address, contract=None):
        """a function to create and pass back a bath token instance

        :param token_address: the address of the token
        :type token_address: str
        :param contract: a bath token contract instance, defaults to None
        :type contract: class, optional
        :return: a bath token instance
        :rtype: bath.Bath
        """

        if self.wallet and self.key:
            return contracts.BathTokenSigner(self.w3, token_address, self.wallet, self.key, contract=contract)
        else:
            return contracts.BathToken(self.w3, token_address, contract=contract)

    # a function to create and pass back an ERC20 token instance
    # TODO: add support for a contract instance
    def token(self, token_address):
        """a function to create and pass back an ERC20 token instance

        :param token_address: the address of the token
        :type token_address: str
        :return: an ERC20 token instance
        :rtype: token.Token
        """

        if self.wallet and self.key:
            return contracts.helper.ERC20Signer(self.w3, token_address, self.wallet, self.key, contract=None)
        else:
            return contracts.helper.ERC20(self.w3, token_address, contract=None)

    # a function to create and pass back a market aid instance
    def aid(self, address, contract=None):
        """a function to create and pass back a market aid instance

        :param address: the address of the market aid contract
        :type address: str
        :param contract: a market aid contract instance, defaults to None
        :type contract: class, optional
        :return: a market aid instance
        :rtype: aid.Aid
        """

        if self.wallet and self.key:
            return contracts.MarketAidSigner(self.w3, address, self.wallet, self.key, contract=contract)
        else:
            return contracts.MarketAid(self.w3, address, contract=contract)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rubi', 'rubi.book', 'rubi.contracts', 'rubi.contracts.helper']

package_data = \
{'': ['*'], 'rubi.contracts.helper': ['abis/*']}

install_requires = \
['attributedict>=0.3.0,<0.4.0',
 'eth-abi>=3.0.1,<4.0.0',
 'eth-tester>=0.8.0b1,<0.9.0',
 'eth-utils>=2.1.0,<3.0.0',
 'hexbytes>=0.3.0,<0.4.0',
 'py-evm>=0.6.1a1,<0.7.0',
 'pytest>=7.2.0,<8.0.0',
 'sphinx>=5.3.0,<6.0.0',
 'web3>5.23.0,<7.0']

setup_kwargs = {
    'name': 'rubi',
    'version': '1.0.0',
    'description': 'a python SDK for the Rubicon Protocol',
    'long_description': '# rubi\nRubicon Python SDK\n\n## Risk Disclaimers\n\n### SDK Disclaimer\n\nThis codebase is in Alpha and could contain bugs or change significantly between versions. Contributing through Issues or Pull Requests is welcome!\n\n### Protocol Disclaimer\n\nPlease refer to [this](https://docs.rubicon.finance/docs/protocol/rubicon-pools/risks) for information on the risks associated to the Rubicon Protocol.',
    'author': 'denver',
    'author_email': 'denver@rubicon.finance',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import hexbytes
from web3 import Web3 
import logging as log
from eth_abi.codec import ABICodec
from web3._utils.events import get_event_data

from rubi.contracts.helper import networks

class BathHouse: 
    """this class represents the BathHouse.sol contract and has read functionality for the contract
    
    :param w3: Web3 instance
    :type w3: Web3
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 object, optional
    """

    # init function
    def __init__(self, w3, contract=None):
        """constructor method"""

        chain = w3.eth.chain_id

        if contract:
            self.contract = contract
            self.address = self.contract.address
        else:
            network = networks[chain]()
            self.contract = w3.eth.contract(address=network.house, abi=network.house_abi)
            self.address = network.house

        self.chain = chain
        self.w3 = w3
        self.log_new_bath_token_abi = self.contract.events.LogNewBathToken()._get_event_abi()
        self.codec: ABICodec = w3.codec

    ######################################################################
    # read calls
    ######################################################################  
    
    # RubiconMarketAddress()
    def rubicon_market_address(self):
        """returns the address of the RubiconMarket contract

        :return: rubicon market address
        :rtype: str
        """
        
        try: 
            rubicon_market_address = self.contract.functions.RubiconMarketAddress().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return rubicon_market_address

    # admin()
    def admin(self):
        """returns the admin address of the contract

        :return: admin address
        :rtype: str
        """

        try: 
            admin = self.contract.functions.admin().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return admin

    # getBathTokenFromAsset(asset (address))
    def get_bath_token_from_asset(self, asset):
        """returns the bathtoken that is associated with the asset, returns a zero address if no bathtoken is associated with the asset

        :param asset: asset address
        :type asset: str
        :return: bath token address
        :rtype: str
        """

        try: 
            bath_token = self.contract.functions.getBathTokenFromAsset(asset).call()
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            bath_token = self.contract.functions.getBathTokenFromAsset(Web3.to_checksum_address(asset)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return bath_token
    
    # newBathTokenImplementation()
    def new_bath_token_implementation(self):
        """returns the newest bath token implementation address

        :return: address of the newest bath token implementation
        :rtype: str
        """

        try: 
            new_bath_token_implementation = self.contract.functions.newBathTokenImplementation().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None
        
        return new_bath_token_implementation

    ######################################################################
    # events & helpers
    ######################################################################

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_new_bath_token_hash(self):
        return self.w3.keccak(text="LogNewBathToken(address,address,address,uint256,address)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_new_bath_token(self, data): 

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_new_bath_token_abi, data['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            new_bath_token = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'underlying_token': event['args']['underlyingToken'],  
                    'bath_token_address': event['args']['bathTokenAddress'],
                    'fee_admin': event['args']['bathTokenFeeAdmin'],
                    'timestamp': event['args']['timestamp'],
                    'creator': event['args']['bathTokenCreator']
            }
            return new_bath_token

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_new_bath_token(self, log): 

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_new_bath_token_abi, log)

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            new_bath_token = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'underlying_token': event['args']['underlyingToken'],  
                    'bath_token_address': event['args']['bathTokenAddress'],
                    'fee_admin': event['args']['bathTokenFeeAdmin'],
                    'timestamp': event['args']['timestamp'],
                    'creator': event['args']['bathTokenCreator']
            }
            return new_bath_token
        
        except Exception as e:
            log.error(e, exc_info=True)
            return None

class BathHouseSigner(BathHouse):
    """this class represents the BathHouse.sol contract and has read and write functionality for the contract. this class inherits from the BathHouse class and adds the ability to sign transactions.
    
    :param w3: Web3 instance
    :type w3: Web3
    :parm wallet: the signers wallet address
    :type wallet: str
    :param key: the signers private key
    :type key: str
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 object, optional
    """

    # init function
    def __init__(self, w3, wallet, key, contract=None):
        super().__init__(w3, contract)
        self.wallet = wallet
        self.key = key
    
    ######################################################################
    # write calls
    ######################################################################

    # createBathToken(underlyingERC20 (address), _feeAdmin (address))
    def create_bath_token(self, underlying_erc20, fee_admin,  nonce=None, gas=3000000, gas_price=None):
        """creates a new bath token

        :param underlying_erc20: address of the underlying token
        :type underlying_erc20: str
        :param fee_admin: address of the fee admin
        :type fee_admin: str
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the create bath token transaction, None if the transaction failed
        :rtype: dict, None
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        if gas_price is None:
            gas_price = self.w3.eth.gas_price
        
        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': nonce}

        try:
            bath_token = self.contract.functions.createBathToken(underlying_erc20, fee_admin).buildTransaction(txn)
            bath_token = self.w3.eth.account.sign_transaction(bath_token, self.wallet.privateKey)
        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            bath_token = self.contract.functions.createBathToken(Web3.to_checksum_address(underlying_erc20), Web3.to_checksum_address(fee_admin)).buildTransaction(txn)
            bath_token = self.w3.eth.account.sign_transaction(bath_token, self.wallet.privateKey)
            self.w3.eth.send_raw_transaction(bath_token.rawTransaction)
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return bath_token
